/*******************************************************************************
 * @file                segBot.c
 * @brief               Motor controller and tachometer
 * @date                October 25th, 2014
 * @author              av333, kb476, avh34
 * @version             1.0
 *
 * This program uses the TRT operating system to schedule three main tasks
 *
 ******************************************************************************/


/*** Program Macros ***********************************************************/

/* Semaphore to protect shared variable */
#define SEM_RX_ISR_SIGNAL 1
#define SEM_STRING_DONE   2


/*** Include Files ************************************************************/

#include "../lib/TRT/trtSettings.h"
#include "../lib/TRT/trtKernel_644.c"

#include <inttypes.h>
#include <stdio.h>
#include <string.h>
#include <avr/io.h>
#include <avr/sleep.h>
#include <util/delay.h>

#include "../lib/wiiNunchuk/wiiNunchuk_lib.h"
#include "../lib/wiiNunchuk/wiiNunchuk_lib.c"


/*** Program Variables ********************************************************/

/* Input arguments to each thread, not actually used. */
int args[3];

float p = 1.6;   /* PID proportional gain */
float i = 0.3;   /* PID integral gain     */
float d = 0.2;   /* PID differential gain */
float s = 0;     /* Desired fan speed     */
float err = 0;


/*** Program Function Prototypes **********************************************/

int  main       (void);
void initialize (void);

void motorTask  (void *args);


/*** Program Functions ********************************************************/

int main(void)
{
	initialize();

	/* Sets up the kernel data structures. The parameter is the desired starck
	 * size of the idle task. For a null idle task, a stack size of 80 should be
	 * sufficient. */
	trtInitKernel(80);

	/* Creates semaphores with identifier semnbr and initial values initval. */
	trtCreateSemaphore(SEM_RX_ISR_SIGNAL, 0);
	trtCreateSemaphore(SEM_STRING_DONE,   0);

	/* Identifies the three functions to the kernel as threads. The parameters
	 * specify pointers to the functions, the desired stack size, the initial
	 * release time, the initial deadline time, and an abitrary data input
	 * structure. */
	trtCreateTask(motorTask, 200, SECONDS2TICKS(0.3), SECONDS2TICKS(0.32), &(args[2]));

	/* Choose our preferred sleep mode */
	set_sleep_mode(SLEEP_MODE_IDLE);

	/* Set sleep enable (SE) bit */
	sleep_enable();

	/* Sleep the CPU when a task isn't running. */
	while (1) { sleep_cpu(); }
}

void initialize(void)
{
	/* Prescaler = 1024*/
	TCCR0B = 5;

	/* Toggle OC0A on compare match and turn on comp-match */
	TCCR0A = (1 << COM0A1) | (1 << WGM00) | (1 << WGM01);
	DDRB = (1 << PINB3);

	DDRA = (1 << PINA0) | (1 << PINA1);

	/* Prescaler = 1024 */
	TCCR2B = 7;

	/* Toggle OC0A on compare match and turn on comp-match */
	TCCR2A = (1 << COM2A1) | (1 << WGM20) | (1 << WGM21);
	DDRD = (1 << PIND5) | (1 << PIND6) | (1 << PIND7);

	PORTA = 0x00;
	PORTD = 0x00;
}

void motorTask(void *args)
{
	uint32_t rel, dead;

	float pErr = 0;	 /* Proportional gain error term. */
	float iErr = 0;	 /* Integral gain error term.     */
	float dErr = 0;	 /* Differential gain error term. */
	float output;	 /* Summation of the three aforementioned error terms. */

	int16_t outX = 0;
	int16_t outY = 0;
	int16_t outZ = 0;

	int joyX = 0;
	int joyY = 0;

	i2c_init();

	/* Single byte write to set the accelerometer to active mode */
	i2c_start_wait(MMA8452Q_ADDR | I2C_WRITE);
	i2c_write(0x2A);
	i2c_write(0x01);
	i2c_stop();

	_delay_ms(5);

	wiiNunchuk_init();

	while(1)
	{
		/* Start to read accelerometer data, start at 0x01, following the multi-
		 * byte protocol found on pg. 17 of the MMA8452Q data sheet */
		i2c_start_wait(MMA8452Q_ADDR | I2C_WRITE);
		i2c_write(0x01);

		i2c_rep_start(MMA8452Q_ADDR | I2C_READ);

		outX  = (i2c_readAck() << 8);
		outX |= (i2c_readAck());
		outX  = (outX >> 4);

		outY  = (i2c_readAck() << 8);
		outY |= (i2c_readAck());
		outY  = (outY >> 4);

		outZ  = (i2c_readAck() << 8);
		outZ |= (i2c_readNak());
		outZ  = (outZ >> 4);

		i2c_stop();

		_delay_ms(5);

		wiiNunchuk_update();

		joyX = wiiNunchuk_getJoyX();
		joyY = wiiNunchuk_getJoyY();

		/* Calculation of the deviation between the desired and actual fan speed. */
		err = s - outX;

		dErr  = err - pErr;

		iErr += err;

		/* Summation of the three aforementioned error terms. */
		output = p*err + i*iErr + d*dErr;

		/* PWM will drive the motor at 12V, which is full speed. */
		if (output > 255) {
			OCR0A = 255;
			OCR2A = 255;

			PORTA = (1 << PINA0);
			PORTD = (1 << PIND5);
		}
		/* PWM will drive the motor at 12V, which is full speed. */
		else if (output < -255) {
			OCR0A = 255;
			OCR2A = 255;

			PORTA = (1 << PINA1);
			PORTD = (1 << PIND6);
		}
		else if (output > 0) {
			OCR0A = (unsigned char) output;
			OCR2A = (unsigned char) output;

			PORTA = (1 << PINA0);
			PORTD = (1 << PIND5);
		}
		else if (output < 0) {
			OCR0A = (unsigned char) output;
			OCR2A = (unsigned char) output;

			PORTA = (1 << PINA1);
			PORTD = (1 << PIND6);
		}

		if (joyY > 180) {
			s = 20;
		}
		else if (joyY > 180) {
			s = -20;
		}
		else {
			s =0;
		}

		pErr = err; // Save current error as previous

		rel  = trtCurrentTime() + SECONDS2TICKS(0.01);
		dead = trtCurrentTime() + SECONDS2TICKS(0.01);
		trtSleepUntil(rel, dead);
	}
}
