/*******************************************************************************
 * @file                I2C_lib.c
 * @brief               I2C master library using hardware TWI interface
 * @date                December 3rd, 2014
 * @authors             Alexander Volkov <av333@cornell.edu>,
 *                      Kevin Bonett     <kb476@cornell.edu>,
 *                      Andre Heil       <avh34@cornell.edu>,
 *                      Peter Fleury     <pfleury@gmx.ch>
 * @version             2.0
 *
 * Implementation of the I2C library. Contains functions to initialize I2C,
 * start and stop transmissions as well as read and write.
 *
 * Originally created by Peter Fleury, then used and formatted by Andre Heil.
 *
 ******************************************************************************/


/*** Program Macros ***********************************************************/

/* Define CPU frequency in MHz here if not defined in the Makefile */
#ifndef F_CPU
#define F_CPU 16000000UL
#endif

#define SCL_CLOCK 100000UL


/*** Include Files ************************************************************/

#include <inttypes.h>
#include <compat/twi.h>

#include "I2C_lib.h"


/*** Program Functions ********************************************************/

/* i2c_init:
 * Initialization of the I2C bus interface. Need to be called only once.
 */
void i2c_init(void)
{
	/* Initialize TWI clock: 100 kHz clock, TWPS = 0 => prescaler = 1 */
	TWSR = 0;                         /* No prescaler */
	TWBR = ((F_CPU/SCL_CLOCK)-16)/2;  /* Must be > 10 for stable operation */

}


/* i2c_start:
 * Issues a start condition and sends address and transfer direction.
 *     Return: 0: Device accessible,
 *             1: Failed to access device
 */
unsigned char i2c_start(unsigned char address)
{
	uint8_t twst;

	/* Send START condition */
	TWCR = (1 << TWINT) | (1 << TWSTA) | (1 << TWEN);

	/* Wait until transmission completed */
	while (!(TWCR & (1 << TWINT)));

	/* Check value of TWI Status Register. Mask prescaler bits. */
	twst = TW_STATUS & 0xF8;
	if ((twst != TW_START) && (twst != TW_REP_START)) { return 1; }

	/* Send device address */
	TWDR = address;
	TWCR = (1 << TWINT) | (1 << TWEN);

	/* Wait until transmission completed and ACK/NACK has been received */
	while (!(TWCR & (1 << TWINT)));

	/* Check value of TWI Status Register. Mask prescaler bits. */
	twst = TW_STATUS & 0xF8;
	if ((twst != TW_MT_SLA_ACK) && (twst != TW_MR_SLA_ACK)) { return 1; }

	return 0;
}


/* i2c_start_wait:
 * Issues a start condition and sends address and transfer direction. If device
 * is busy, use ack polling to wait until device is ready.
 *     Input: address and transfer direction of I2C device
 */
void i2c_start_wait(unsigned char address)
{
	uint8_t twst;

	while (1) {
		/* Send START condition */
		TWCR = (1 << TWINT) | (1 << TWSTA) | (1 << TWEN);

		/* Wait until transmission completed */
		while (!(TWCR & (1 << TWINT)));

		/* Check value of TWI Status Register. Mask prescaler bits. */
		twst = TW_STATUS & 0xF8;
		if ((twst != TW_START) && (twst != TW_REP_START)) { continue; }

		/* Send device address */
		TWDR = address;
		TWCR = (1 << TWINT) | (1 << TWEN);

		/* Wait until transmission completed */
		while (!(TWCR & (1 << TWINT)));

		/* Check value of TWI Status Register. Mask prescaler bits. */
		twst = TW_STATUS & 0xF8;
		if ((twst == TW_MT_SLA_NACK) || (twst == TW_MR_DATA_NACK))
		{
			/* Device busy, send stop condition to terminate write operation */
			TWCR = (1 << TWINT) | (1 << TWEN) | (1 << TWSTO);

			/* Wait until stop condition is executed and bus released */
			while (TWCR & (1 << TWSTO));

			continue;
		}
		break;
	}
}


/* i2c_rep_start:
 * Issues a repeated start condition and sends address and transfer direction.
 *     Input:   Address and transfer direction of I2C device
 *     Return:  0: Device accessible
 *              1: Failed to access device
 */
unsigned char i2c_rep_start(unsigned char address)
{
	return i2c_start(address);
}


/* i2c_stop:
 * Terminates the data transfer and releases the I2C bus.
 */
void i2c_stop(void)
{
	/* Send stop condition */
	TWCR = (1 << TWINT) | (1 << TWEN) | (1 << TWSTO);

	/* Wait until stop condition is executed and bus released */
	while (TWCR & (1 << TWSTO));
}


/* i2c_write:
 * Send one byte to I2C device.
 *     Input:    Byte to be transfered
 *     Return:   0: Write successful
 *               1: Write failed
 */
unsigned char i2c_write(unsigned char data)
{
    uint8_t twst;

	/* Send data to the previously addressed device */
	TWDR = data;
	TWCR = (1 << TWINT) | (1 << TWEN);

	/* Wait until transmission completed */
	while (!(TWCR & (1 << TWINT)));

	/* Check value of TWI Status Register. Mask prescaler bits. */
	twst = TW_STATUS & 0xF8;
	if (twst != TW_MT_DATA_ACK) { return 1; }

	return 0;
}


/* i2c_readAck:
 * Read one byte from the I2C device, request more data from device.
 *     Return: Byte read from I2C device
 */
unsigned char i2c_readAck(void)
{
	TWCR = (1 << TWINT) | (1 << TWEN) | (1 << TWEA);
	while (!(TWCR & (1 << TWINT)));

	return TWDR;
}


/* i2c_readNak:
 * Read one byte from the I2C device, read is followed by a stop condition
 *     Return: Byte read from I2C device
 */
unsigned char i2c_readNak(void)
{
	TWCR = (1 << TWINT) | (1 << TWEN);
	while (!(TWCR & (1 << TWINT)));

	return TWDR;
}
