/*******************************************************************************
 * @file                I2C_lib.h
 * @brief               I2C (TWI) Master Software Library
 * @date                December 3rd, 2014
 * @authors             Alexander Volkov <av333@cornell.edu>,
 *                      Kevin Bonett     <kb476@cornell.edu>,
 *                      Andre Heil       <avh34@cornell.edu>,
 *                      Peter Fleury     <pfleury@gmx.ch>
 * @version             2.0
 *
 * Basic routines for communicating with I2C slave devices. This single master
 * implementation is limited to one bus master on the I2C bus.
 *
 * This I2C library is implemented as a compact assembler software
 * implementation of the I2C protocol which runs on any AVR (i2cMaster.S) and as
 * a TWI hardware interface for all AVR with built-in TWI hardware
 * (i2cMaster.c). Since the API for these two implementations is exactly the
 * same, an application can be linked either against the software I2C
 * implementation or the hardware I2C implementation.
 *
 * Use 4.7k pull-up resistor on the SDA and SCL pin.
 *
 * Adapt the SCL and SDA port and pin definitions and eventually the delay
 * routine in the module i2cMaster.S to your target when using the software I2C
 * implementation !
 *
 * Adjust the CPU clock frequency F_CPU in i2cMaster.c or in the Makefile when
 * using the TWI hardware implementaion.
 *
 *     The module i2cmaster.S is based on the Atmel Application Note AVR300,
 *     corrected and adapted to GNU assembler and AVR-GCC C call interface.
 *     Replaced the incorrect quarter period delays found in AVR300 with half
 *     period delays.
 *
 * @code
 * // Device address of EEPROM 24C02, see datasheet
 * #define DEV24C02  0xA2
 *
 * #include <i2cmaster.h>
 *
 * int main(void)
 * {
 *     unsigned char ret;
 *
 *     i2c_init();
 *
 *     // Write 0x75 to EEPROM address 5 (Byte Write)
 *     i2c_start_wait(DEV24C02+I2C_WRITE);  // Set device address and write mode
 *     i2c_write(0x05);                     // Write address = 5
 *     i2c_write(0x75);                     // Write value 0x75 to EEPROM
 *     i2c_stop();                          // Set stop conditon = release bus
 *
 *
 *     // Read previously written value back from EEPROM address 5
 *     i2c_start_wait(DEV24C02+I2C_WRITE);  // Set device address and write mode
 *
 *     i2c_write(0x05);                     // Write address = 5
 *     i2c_rep_start(DEV24C02+I2C_READ);    // Set device address and read mode
 *
 *     ret = i2c_readNak();                 // Read one byte from EEPROM
 *     i2c_stop();
 *
 *     for(;;);
 * }
 * @endcode
 ******************************************************************************/

#ifndef _I2CMASTER_H
#define _I2CMASTER_H 1


/*** Program Macros ***********************************************************/

#if (__GNUC__ * 100 + __GNUC_MINOR__) < 304
#error "This library requires AVR-GCC 3.4 or later, update AVR-GCC compiler !"
#endif

/* Defines the data direction (reading from I2C device) in i2c_start() */
#define I2C_READ    1

/* Defines the data direction (writing to   I2C device) in i2c_start() */
#define I2C_WRITE   0


/*** Include Files ************************************************************/

#include <avr/io.h>


/*** Program Function Prototypes **********************************************/

/* i2c_init:
 * @brief  Initialize the I2C master interace. Need to be called only once.
 * @param  void
 * @return none
 */
extern void i2c_init(void);

/* i2c_stop:
 * @brief    Terminates the data transfer and releases the I2C bus.
 * @param    void
 * @return   none
 */
extern void i2c_stop(void);

/* i2c_start:
 * @brief    Issues a start condition and sends address and transfer direction.
 *
 * @param    addr Address and transfer direction of I2C device
 * @retval   0    Device accessible
 * @retval   1    Failed to access device
 */
extern unsigned char i2c_start(unsigned char addr);

/* i2c_rep_start:
 * @brief   Issues a repeated start condition and sends address and transfer.
 * direction.
 *
 * @param   addr Address and transfer direction of I2C device
 * @retval  0    Device accessible
 * @retval  1    Failed to access device
 */
extern unsigned char i2c_rep_start(unsigned char addr);

/* i2c_start_wait:
 * @brief    Issues a start condition and sends address and transfer direction.
 * If the device is busy, use ack polling to wait until device ready
 *
 * @param    addr Address and transfer direction of I2C device
 * @return   none
 */
extern void i2c_start_wait(unsigned char addr);

/* i2c_write:
 * @brief    Send one byte to I2C device
 *
 * @param    data  byte to be transfered
 * @retval   0 write successful
 * @retval   1 write failed
 */
extern unsigned char i2c_write(unsigned char data);


/* i2c_readAck:
 * @brief    Read one byte from the I2C device, request more data from device.
 *
 * @return   byte Read from I2C device
 */
extern unsigned char i2c_readAck(void);

/* i2c_readNak:
 * @brief    Read one byte from the I2C device, read is followed by a stop
 * condition.
 *
 * @return   byte Read from I2C device
 */
extern unsigned char i2c_readNak(void);

/* i2c_read:
 * @brief    Read one byte from the I2C device. Implemented as a macro, which
 * calls either i2c_readAck or i2c_readNak
 *
 * @param    ack  1 Send ack, request more data from device
 *                0 Send nak, read is followed by a stop condition
 * @return   byte   Read from I2C device
 */
extern unsigned char i2c_read(unsigned char ack);
#define i2c_read(ack)  (ack) ? i2c_readAck() : i2c_readNak();

#endif
