/*******************************************************************************
 * @file                nRF24L01_lib.c
 * @brief               Library of functions used to control the nRF24L01+
 * @date                December 3rd, 2014
 * @authors             Alexander Volkov <av333@cornell.edu>,
 *                      Kevin Bonett     <kb476@cornell.edu>,
 *                      Andre Heil       <avh34@cornell.edu>,
 *                      Yi Heng Lee,
 *                      Jaehyun Kim
 * @version             2.0
 *
 * Implementation of the nRF24L01+ library. Contains functions to initialize the
 * chip, transmit, receive, and any basic radio operation.
 *
 * Originally created by Jaehyun Kim, then built upon by Yi Heng Lee, then
 * used and formatted by Andre Heil.
 *
 ******************************************************************************/

/*** Include Files ************************************************************/

#include "nRF24L01_cmdList.h"
#include "nRF24L01_regMap.h"
#include "nRF24L01_lib.h"

#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdio.h>
#include <util/delay.h>


/*** Program Functions ********************************************************/

/* init_RF:
 * Initialize the nRF24L01+ transceiver. Set en_aa to 0 to disable enhanced
 * shockburst and 1 to enable. The transceiver starts out in RX mode.
 */
void init_RF(unsigned char en_aa)
{
	if (PAYLOAD_LENGTH > 32 || PAYLOAD_LENGTH < 1) {
		fprintf(stdout,"PAYLOAD_LENGTH must be in range 1 to 32\n");
		return;
	}

	/* SPI settings for ATmega644 */
	DDR_SPI = (1 << DD_MOSI) | (1 << DD_SCK) | (1 << DD_CSN) | (1 << DD_CE);
	SPCR    = (1 << SPE) | (1 << MSTR);

	/* Delay is needed for module to power on */
	_set_CE(0);
	_delay_ms(100);

	/* Setup transceiver registers:
	 * Address width is set to 4 bytes, RF channel is set to 0x00, data rate is
	 * set to 2 MBps and the output power is set to 0 dBm, and the payload
	 * length is set for data pipes 1 and 2.
	 */
	_write_RF_register(SETUP_AW, 0x02);
	_write_RF_register(RF_CH,    0x00);
	_write_RF_register(RF_SETUP, (1 << RF_DR_HIGH) | (0x3 << RF_PWR));
	_write_RF_register(RX_PW_P0, PAYLOAD_LENGTH);
	_write_RF_register(RX_PW_P1, PAYLOAD_LENGTH);
	_write_RF_register(FEATURE,  (0 << EN_DPL) | (0<<EN_ACK_PAY);
	_write_RF_register(DYNPD,    0x00);

	/* Enable or disable enhanced shockburst on all data pipes */
	if (en_aa == 1) {
		_write_RF_register(EN_AA, 0x3F);
	}
	else {
		_write_RF_register(EN_AA, 0x00);
	}

	_write_RF_register(CONFIG, (1 << EN_CRC) | (1 << PWR_UP) | (1 << PRIM_RX));

	/* Clear IRQs and FIFOs */
	_clear_IRQs();
	_flush_TX_FIFO();
	_flush_RX_FIFO();

	/* Delay for module to power up */
	_delay_ms(5);
	_set_CE(1);
}

/* enable_pipe:
 * Enables the pipe specified by pipeNumber, which can be from 0 to 5.
 */
void enable_pipe(unsigned char pipeNumber)
{
	if (pipeNumber < 0 || pipeNumber > 5) {
		fprintf(stdout,"Pipe number must be in range 0 to 5\n");
		return;
	}

	if (PAYLOAD_LENGTH > 32 || PAYLOAD_LENGTH < 1) {
		fprintf(stdout,"PAYLOAD_LENGTH must be in range 1 to 32\n");
		return;
	}

	unsigned char en_rxaddr = _read_RF_register(EN_RXADDR);

	switch (pipeNumber) {
		case 0:
			en_rxaddr |= (1 << ERX_P0);
			_write_RF_register(RX_PW_P0, PAYLOAD_LENGTH);
			break;
		case 1:
			en_rxaddr |= (1 << ERX_P1);
			_write_RF_register(RX_PW_P1, PAYLOAD_LENGTH);
			break;
		case 2:
			en_rxaddr |= (1 << ERX_P2);
			_write_RF_register(RX_PW_P2, PAYLOAD_LENGTH);
			break;
		case 3:
			en_rxaddr |= (1 << ERX_P3);
			_write_RF_register(RX_PW_P3, PAYLOAD_LENGTH);
			break;
		case 4:
			en_rxaddr |= (1 << ERX_P4);
			_write_RF_register(RX_PW_P4, PAYLOAD_LENGTH);
			break;
		case 5:
			en_rxaddr |= (1 << ERX_P5);
			_write_RF_register(RX_PW_P5, PAYLOAD_LENGTH);
			break;
	}

	_write_RF_register(EN_RXADDR, en_rxaddr);
}

/* set_mode:
 * Sets the transciever to TX mode if mode == TX_MODE, else set it to RX mode
 */
void set_mode(unsigned char mode)
{
	unsigned char config = _read_RF_register(CONFIG);

	if (mode == TX_MODE) {
		/* CE is 0 in TX mode except when transmitting */
		_set_CE(0);

		config &= ~(1 << PRIM_RX);
		_write_RF_register(CONFIG, config);

		_delay_us(130);
	}
	else {
		config |= 1<<PRIM_RX;
		_write_RF_register(CONFIG, config);

		/* CE is kept at 1 in RX mode */
		_set_CE(1);

		_delay_us(130);
	}
}

/* transmit_packet:
 * Transmit up to PAYLOAD_LENGTH bytes of the packet.
 *     return: 1 if the transmission is successful,
 *             2 if maximum number of retries reached without receiving ACK.
 *             This only occurs if enhanced shockburst is enabled.
 * With enhanced shockburst disabled, this function always returns 1 unless the
 * transciever is connected incorrectly.
 */
unsigned char transmit_packet(char* packet)
{
	unsigned char i;
	char TX_status;

	/* Start sending packet to TX FIFO */
	_set_CSN(0);

	status = _SPI_transmit(W_TX_PAYLOAD);
	for (i = 0; i < PAYLOAD_LENGTH; i++) {
		_SPI_transmit(packet[i]);
	}

	_set_CSN(1);

	/* Set CE to 1 to begin transmission */
	_set_CE(1);

	/* CE must be high for at least 10 us to switch to the TX settling state */
	_delay_us(10);

	/* 130 us in the TX settling state */
	_delay_us(130);

	/* Loop until transmission successful or MAX_RT high */
	while (1) {
		TX_status = _check_TX_status();
		if (TX_status == 1) { return 1; }
		if (TX_status == 2) { return 2; }
	}

	_set_CE(0);
}

/* RX_FIFO_status:
 *     return 0 - RX_FIFO is empty
 *            1 - Otherwise
 */
unsigned char RX_FIFO_status(void)
{
	_read_status();

	if ((status & 0xE) >> 1 == 7) { return 0; }
	else                          { return 1; }
}

/* read_RX_payload:
 * Reads up to PAYLOAD_LENGTH bytes of a received packet from the RX FIFO into
 * buffer.
 *     return: The pipe number that the packet was received from.
 */
unsigned char read_RX_payload(char* buffer)
{
	unsigned char i;

	_set_CE(0);
	_set_CSN(0);

	_SPI_transmit(R_RX_PAYLOAD);

	status = SPDR;

	for (i = 0; i < PAYLOAD_LENGTH; i++){
		_SPI_transmit(0x00);
		buffer[i] = SPDR;
	}

	/* End transmission */
	_set_CSN(1);
	_set_CE(1);

	/* Returns RX_P_NO in the STATUS register, which indicates the pipe that the
	 * payload was received from, RX_P_NO = 6 indicates not used and RX_P_NO = 7
	 * indicates RX FIFO empty */
	return (status & 0xE) >> 1;
}

/* set_TX_addr:
 * Sets the TX address of the transceiver.
 */
void set_TX_addr(unsigned long address)
{
	_write_RF_register_long(TX_ADDR, address);
}

/* set_RX_addr:
 * Sets the RX address for the specified pipe for transmitter, RX address for
 * pipe 0 should be the same as the TX address.
 */
void set_RX_addr(unsigned long address, unsigned char pipe)
{
	if (pipe == 0 || pipe == 1) {
		_write_RF_register_long(RX_ADDR_P0 + pipe, address);
	}
	else {
		unsigned char temp8 = address;
		_write_RF_register(RX_ADDR_P0 + pipe, temp8);
	}
}

/* set_RF_channel:
 * Sets the RF channel to channel, channel is a 7 bit number
 */
void set_RF_channel(unsigned char channel)
{
	_write_RF_register(RF_CH, channel);
}


/*** Internal Helper Functions ************************************************/

/* _set_CSN:
 * CSN pin is used to initiate and end the communication between the SPI module
 * of ATmega644 and the RF chip.
 *     value = 1 - Start SPI communication
 *     value = 0 - End   SPI communication
 */
void _set_CSN(unsigned char value)
{
	if      (value == 1) { PORTB |=  (1 << CSN); }
	else if (value == 0) { PORTB &= ~(1 << CSN); }
}

/* _set_CE:
 * CE enables and disables the transceiver from transmitting and receiving.
 *     value = 1 - Enables
 *     value = 0 - Disables
 */
void _set_CE(unsigned char value)
{
	if      (value == 1) { PORTB |=  (1 << CE); }
	else if (value == 0) { PORTB &= ~(1 << CE); }
}

/* _SPI_transmit:
 * Helper function for transmitting data by SPI
 */
unsigned char _SPI_transmit(unsigned char data)
{
	SPDR = data;

	/* Wait for the data to be transmitted, bit 7 of SPSR is set high when
	 *serial transfer is complete */
	while(!(SPSR & (1 << SPIF)));

	/* Flag is cleared by reading SPDR */
	return SPDR;
}

/* _read_RF_register:
 * Reads the contents of the register specified by addr
 */
unsigned char _read_RF_register(unsigned char addr)
{
	unsigned char temp8;

	if((addr > 0x17 && addr < 0x1C) || addr > 0x1D) {
		return 0xFFFF;
	}

	_set_CSN(0);

	temp8 = (R_REGISTER | addr);
	_SPI_transmit(temp8);

	status = SPDR;
	_SPI_transmit(0x00);

	_set_CSN(1);
	temp8 = SPDR;

	return temp8;
}

/* _write_RF_register:
 * Writes unsigned char value to the register specified by addr.
 */
void _write_RF_register(unsigned char addr, unsigned char value)
{
	unsigned char temp8;

	_set_CSN(0);

	temp8 = (W_REGISTER | addr);

	_SPI_transmit(temp8);
	_SPI_transmit(value);

	_set_CSN(1);
}

/* _write_RF_register_long:
 * Writes unsigned long longvalue to the register specified by addr.
 */
void _write_RF_register_long(unsigned char addr, unsigned char longvalue)
{
	unsigned char temp8;

	_set_CSN(0);

	temp8 = (W_REGISTER | addr);
	_SPI_transmit(temp8);
	status = SPDR;

	temp8 = longvalue;
	_SPI_transmit(temp8);

	temp8 = (longvalue >> 8);
	_SPI_transmit(temp8);

	temp8 = (longvalue >> 16);
	_SPI_transmit(temp8);

	temp8 = (longvalue >> 24);
	_SPI_transmit(temp8);

	_set_CSN(1);
}

/* _clear_IRQs:
 * Clear the RX_DR, TX_DS and MAX_RT IRQs by writing to 1 to bits corresponding
 * to them in the status register.
 */
void _clear_IRQs(void)
{
	_write_RF_register(STATUS, (1 << RX_DR) | (1 << TX_DS) | (1 << MAX_RT));
}

/* _check_TX_status:
 *     return 0 - Tranmission is still in progress
 *            1 - TX_DS is high meaning that ACK was received and packet was
 *                transmitted successfully
 *            2 - MAX_RT is high meaning that the max number of retries has been
 *                reached
 */
unsigned char _check_TX_status(void)
{
	_read_status();

	if (status & (1 << TX_DS)){
		_clear_IRQs();
	 	return 1;
	}

	if (status & (1 << MAX_RT)){
		_clear_IRQs();
	 	return 2;
	}

	else {
		return 0;
	}
}

/* _read_status:
 * Returns the value of the status register.
 */
unsigned char _read_status(void)
{
	_set_CSN(0);

	_SPI_transmit(NOP);
	status = SPDR;

	_set_CSN(1);
	return status;
}

/* _flush_TX_FIFO:
 * Flushes the TX_FIFO
 */
void _flush_TX_FIFO(void)
{
	_set_CSN(0);

	_SPI_transmit(FLUSH_TX);
	status = SPDR;

	_set_CSN(1);
}

/* _flush_RX_FIFO:
 * Flushes the RX_FIFO.
 */
void _flush_RX_FIFO(void)
{
	_set_CSN(0);

	_SPI_transmit(FLUSH_RX);
	status = SPDR;

	_set_CSN(1);
}

/* _check_RX_DR:
 * Returns true if the RX_DR flag is set, indicating that a packet was received.
 */
unsigned char _check_RX_DR(void)
{
	_read_status();

	if (status & (1 << RX_DR)) {
		_clear_IRQs();
	 	return 1;
	}

	else {
		return 0;
	}
}
