/*******************************************************************************
 * @file                trtKernel_644.c
 * @brief               The TinyRealTime OS
 * @date                December 2nd, 2014
 * @authors             Alexander Volkov <av333@cornell.edu>,
 *                      Kevin Bonett     <kb476@cornell.edu>,
 *                      Andre Heil       <avh34@cornell.edu>,
 *                      Bruce Land       <brl4@cornell.edu>,
 *                      Anton Cervin     <anton@control.lth.se>,
 *                      Dan Henriksson   <dan@control.lth.se>
 * @version             3.0
 *
 * From http://www.control.lth.se/%7Eanton/tinyrealtime/
 *
 * Modified for the ATmega644 by Bruce Land Jan 2009:
 *     Changed register names to ATmega644 regs
 *     Added one byte to stack init section in trtCreateTask
 *
 * Originally created by Anton Cervin and Dan Henriksson, then built upon by
 * Bruce Land, then used and formatted by Andre Heil.
 *
 ******************************************************************************/


/*** Include Files ************************************************************/

#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>


/*** Program Macros ***********************************************************/

#define TERMINATED  0
#define READYQ      1
#define TIMEQ       2
#define WAIT_OFFSET 2

#define SECONDS2TICKS(T) ((uint32_t)((T)*TICKSPERSECOND))

#define lo8(X) ((uint8_t)((uint16_t)(X)))
#define hi8(X) ((uint8_t)((uint16_t)(X) >> 8))


/*** Kernel Data Structures ***************************************************/

/* task:
 *     spl:   Stack pointer lo8
 *     sph:   Stack pointer hi8
 *     state: 0 = Terminated,
 *            1 = ReadyQ,
 *            2 = TimeQ,
 *            3 = Waiting for Sem1,
 *            4 = Waiting for Sem2, etc.
 */
struct task {
	uint8_t  spl;
	uint8_t  sph;
	uint32_t release;
	uint32_t deadline;
	uint8_t  state;
};

/* kernel:
 *     nbrOfTasks: Number of tasks created so far
 *     tasks:      +1 for the idle task
 *     semaphores: Counters for semaphores
 *     memptr:     Pointer to free memory
 *     cycles:     Number of major cycles since system start
 *     nextHit:    Next kernel wake-up time
 */
struct kernel {
	uint8_t  nbrOfTasks;
	uint8_t  running;
	struct   task tasks[MAXNBRTASKS+1];
	uint8_t  semaphores[MAXNBRSEMAPHORES];
	uint8_t  *memptr;
	uint32_t cycles;
	uint32_t nextHit;
} kernel;


/*** Program Functions ********************************************************/

void trtInitKernel(int idlestack)
{
	/* Set up timer 1: reset counter 1, set to normal operation, prescalar of
	 * 1024, and turn on compare match ISR */
	TCNT1  = 0x0000;
	TCCR1A = 0x00;
	TCCR1B = PRESCALEBITS;
	TIMSK1 = (1 << OCIE1A);

	kernel.memptr = (void *) (RAMEND - idlestack);
	kernel.nbrOfTasks = 0;
	kernel.running = 0;
	kernel.cycles = 0x0000;
	kernel.nextHit = 0x7FFFFFFF;

	/* Initialize idle task (task 0) */
	kernel.tasks[0].deadline = 0x7FFFFFFF;
	kernel.tasks[0].release = 0x00000000;

	sei();
}

void trtCreateTask(void (*fun)(void*), uint16_t stacksize, uint32_t release, uint32_t deadline, void *args)
{
	uint8_t *sp;
	struct task *t;
	int i;

	cli();

	++kernel.nbrOfTasks;

	sp = kernel.memptr;

	/* Decrease free memptr */
	kernel.memptr -= stacksize;

	/* Initialize stack */
	*sp-- = lo8(fun);
	*sp-- = hi8(fun);

	/* Store SREG, R0 - R1, R3 - R23 */
	for (i = 0; i < 26; i++) {
		*sp-- = 0x00;
	}

	/* Save args in R24 - R25 (input arguments stored in these registers) */
	*sp-- = lo8(args);
	*sp-- = hi8(args);

	/* Store R26 - R31 */
	for (i=0; i<6; i++) {
		*sp-- = 0x00;
	}

	t = &kernel.tasks[kernel.nbrOfTasks];

	t->release = release;
	t->deadline = deadline;
	t->state = TIMEQ;

	/* Store stack pointer */
	t->spl = lo8(sp);
	t->sph = hi8(sp);

	/* Call interrupt handler to schedule */
	TIMER1_COMPA_vect();
}

void trtCreateSemaphore(uint8_t semnbr, uint8_t initVal)
{
	cli();
	kernel.semaphores[semnbr-1] = initVal;
	sei();
}

void trtWait(uint8_t semnbr)
{
	struct task *t;
	uint8_t *s;

	t = &kernel.tasks[kernel.running];

	cli();

	s = &kernel.semaphores[semnbr-1];
	if ((*s) > 0) {
		(*s)--;
	}
	else {
		/* Waiting for semnbr */
		t->state = semnbr + WAIT_OFFSET;

		/* Call interrupt handler to schedule */
		TIMER1_COMPA_vect();
	}

	sei();
}

void trtSignal(uint8_t semnbr)
{
	uint8_t i;
	struct task *t;
	uint32_t minDeadline = 0xFFFFFFFF;
	uint8_t taskToReadyQ = 0;

	cli();

	for (i=1; i <= kernel.nbrOfTasks; i++) {
		t = &kernel.tasks[i];

		if (t->state == (semnbr + WAIT_OFFSET)) {
			if (t->deadline <= minDeadline) {
				taskToReadyQ = i;
				minDeadline = t->deadline;
			}
		}
	}

	if (taskToReadyQ == 0) {
		kernel.semaphores[semnbr-1]++;
	}
	else {
		kernel.tasks[taskToReadyQ].state = READYQ;

		/* Call interrupt handler to schedule */
		TIMER1_COMPA_vect();
	}

	sei();
}

uint32_t trtCurrentTime(void)
{
	return (((uint32_t)kernel.cycles << 16) + (uint32_t)TCNT1);
}


void trtSleepUntil(uint32_t release, uint32_t deadline)
{
	struct task *t;

	t = &kernel.tasks[kernel.running];

	cli();

	t->state = TIMEQ;
	t->release = release;
	t->deadline = deadline;

	/* Call interrupt handler to schedule */
	TIMER1_COMPA_vect();
}


uint32_t trtGetRelease(void)
{
	return kernel.tasks[kernel.running].release;
}

uint32_t trtGetDeadline(void)
{
	return kernel.tasks[kernel.running].deadline;
}

void trtTerminate(void)
{
	cli();

	kernel.tasks[kernel.running].state = TERMINATED;

	/* Call interrupt handler to schedule */
	TIMER1_COMPA_vect();
}

uint8_t trtAccept(uint8_t semnbr)
{
	uint8_t *s;
	uint8_t temp;

	cli();

	s = &kernel.semaphores[semnbr-1];
	temp = *s ;

	if ((*s) > 0) {
		(*s)--;
	}

	sei();
	return temp;
}


/*** Interrupt Service Routines ***********************************************/

ISR(TIMER1_COMPA_vect) {

	uint8_t running, oldrunning;
	struct task *t;
	uint8_t i;
	uint32_t now;
	uint32_t nextHit;
	int32_t timeleft;

	TIMSK1 = 0;
	nextHit = 0x7FFFFFFF;
	oldrunning = kernel.running;
	running = 0;

	if (TIFR1 & (1 << TOV1)) {
		++kernel.cycles;
		TIFR1 |= (1 << TOV1) ;
	}

	/* Read clock */
	now = (kernel.cycles << 16) + TCNT1;

	/* Release tasks from TimeQ and determine new running task */
	for (i=1; i <= kernel.nbrOfTasks; i++) {

		t = &kernel.tasks[i];

		if (t->state == TIMEQ) {
			if (t->release <= now) {
				t->state = READYQ;
			}
			else if (t->release < nextHit) {
				nextHit = t->release;
			}
		}

		if (t->state == READYQ) {
			if (t->deadline < kernel.tasks[running].deadline) {
				running = i;
			}
		}

	}

	/* Perform context switch */
	if (running != oldrunning) {

		/* Store old context */
		t = &kernel.tasks[oldrunning];
		t->spl = SPL;
		t->sph = SPH;

		/* Load new context */
		t = &kernel.tasks[running];
		SPH = t->sph;
		SPL = t->spl;

		kernel.running = running;
	}

	kernel.nextHit = nextHit;

	now = (kernel.cycles << 16) + TCNT1;
	timeleft = (int32_t)nextHit - (int32_t)now;
	if (timeleft < 4) {
		timeleft = 4;
	}

	if ((unsigned long)TCNT1 + timeleft < 65536) {
		OCR1A = TCNT1 + timeleft;
	}
	else if (TCNT1 < 65536 - 4) {
		OCR1A = 0x0000;
	}
	else {
		OCR1A = 4;
	}

	TIMSK1 = (1 << OCIE1A);
}
