/*******************************************************************************
 * @file                wiiNunchuk_lib.c
 * @brief               wiiNunchuk library
 * @date                December 2nd, 2014
 * @authors             Alexander Volkov <av333@cornell.edu>,
 *                      Kevin Bonett     <kb476@cornell.edu>,
 *                      Andre Heil       <avh34@cornell.edu>,
 *                      David Gironi
 * @version             2.0
 *
 * Implementation of the wiiNunchuk library. Contains functions to initialize the
 * controller, update, and get controller parameters.
 *
 * Originally created by David Gironi, then used and formatted by Andre Heil.
 *
 ******************************************************************************/


/*** Include Files ************************************************************/

#include <math.h>
#include <stdio.h>
#include <string.h>
#include <avr/io.h>
#include <util/delay.h>

#include "wiiNunchuk_lib.h"
#include "../I2C/I2C_lib.h"


/*** Program Variables ********************************************************/

volatile uint8_t wiiNunchuk_joyX    = 0;
volatile uint8_t wiiNunchuk_joyY    = 0;
volatile uint8_t wiiNunchuk_buttonZ = 0;
volatile uint8_t wiiNunchuk_buttonC = 0;
volatile int     wiiNunchuk_angleX  = 0;
volatile int     wiiNunchuk_angleY  = 0;
volatile int     wiiNunchuk_angleZ  = 0;

#if WIINUNCHUK_PULSEBUTTON == 1
volatile uint8_t wiiNunchuk_lastButtonZ = 0;
volatile uint8_t wiiNunchuk_lastButtonC = 0;
#endif

#if WIINUNCHUK_ANGLEFILTER == 1
volatile int wiiNunchuk_averageAngleX[WIINUNCHUK_ANGLEAVERAGECOEF];
volatile int wiiNunchuk_averageAngleY[WIINUNCHUK_ANGLEAVERAGECOEF];
volatile int wiiNunchuk_averageAngleZ[WIINUNCHUK_ANGLEAVERAGECOEF];
#endif


/*** Program Functions ********************************************************/

int wiiNunchuk_getJoyX()
{
	#if WIINUNCHUK_GETNONCALIBRATED == 1
	return (int) wiiNunchuk_joyX;

	#elif WIINUNCHUK_JOYCENTERB == 1
	int joyX = (int) wiiNunchuk_joyX;
	if(joyX <= WIINUNCHUK_CENTERJOYX)
		joyX = (joyX - WIINUNCHUK_INMINJOYX ) * (127 -   0) / (WIINUNCHUK_CENTERJOYX - WIINUNCHUK_INMINJOYX) + 0;
	else
		joyX = (joyX - WIINUNCHUK_CENTERJOYX) * (255 - 127) / (WIINUNCHUK_INMAXJOYX - WIINUNCHUK_CENTERJOYX) + 127 - 1;
	return joyX - 127;

	#else
	return (int) wiiNunchuk_joyX - WIINUNCHUK_DEFAULTZEROJOYX;

	#endif
}

int wiiNunchuk_getJoyY()
{
	#if WIINUNCHUK_GETNONCALIBRATED == 1
	return (int)wiiNunchuk_joyY;

	#elif WIINUNCHUK_JOYCENTERB == 1
	int joyY = (int)wiiNunchuk_joyY;
	if(joyY <= WIINUNCHUK_CENTERJOYY)
		joyY = (joyY - WIINUNCHUK_INMINJOYY ) * (127 -   0) / (WIINUNCHUK_CENTERJOYY - WIINUNCHUK_INMINJOYY) + 0;
	else
		joyY = (joyY - WIINUNCHUK_CENTERJOYY) * (255 - 127) / (WIINUNCHUK_INMAXJOYY - WIINUNCHUK_CENTERJOYY) + 127 - 1;
	return joyY - 127;

	#else
	return (int)wiiNunchuk_joyY - WIINUNCHUK_DEFAULTZEROJOYY;

	#endif
}

uint8_t wiiNunchuk_getButtonZ()
{
	#if WIINUNCHUK_PULSEBUTTON == 1
	return (wiiNunchuk_buttonZ && !wiiNunchuk_lastButtonZ);

	#else
	return wiiNunchuk_buttonZ;

	#endif
}

uint8_t wiiNunchuk_getButtonC()
{
	#if WIINUNCHUK_PULSEBUTTON == 1
	return (wiiNunchuk_buttonC && !wiiNunchuk_lastButtonC);

	#else
	return wiiNunchuk_buttonC;

	#endif
}

int wiiNunchuk_getAngleX()
{
	#if WIINUNCHUK_GETNONCALIBRATED == 1
	return wiiNunchuk_angleX;

	#else
	return wiiNunchuk_angleX - WIINUNCHUK_ZEROANGLEX;

	#endif
}

int wiiNunchuk_getAngleY()
{
	#if WIINUNCHUK_GETNONCALIBRATED == 1
	return wiiNunchuk_angleY;

	#else
	return wiiNunchuk_angleY - WIINUNCHUK_ZEROANGLEY;

	#endif
}

int wiiNunchuk_getAngleZ()
{
	#if WIINUNCHUK_GETNONCALIBRATED == 1
	return wiiNunchuk_angleZ;

	#else
	return wiiNunchuk_angleZ - WIINUNCHUK_ZEROANGLEZ;

	#endif
}

#if WIINUNCHUK_GETATTITUDE == 1
void wiiNunchuk_getPitchRoll(double ax, double ay, double az, double *pitch, double *roll)
{
	double magnitude = sqrt(((uint32_t)ax * (uint32_t)ax) + ((uint32_t)ay * (uint32_t)ay) + ((uint32_t)az * (uint32_t)az));
	ax = ax / magnitude;
	ay = ay / magnitude;
	az = az / magnitude;
	*pitch = atan2(ax, sqrt(pow(ay,2) + pow(az, 2)));
	*roll  = atan2(ay, az);
}

#endif

#if WIINUNCHUK_ANGLEFILTER == 1
int wiiNunchuk_averageFilter(int input, int wiiNunchuk_averageAngle[])
{
	uint8_t i   = 0;
	long    sum = 0;

	for (i = 0; i < WIINUNCHUK_ANGLEAVERAGECOEF; i++) {
		wiiNunchuk_averageAngle[i] = wiiNunchuk_averageAngle[i + 1];
	}

	wiiNunchuk_averageAngle[WIINUNCHUK_ANGLEAVERAGECOEF - 1] = input;

	for (i = 0; i < WIINUNCHUK_ANGLEAVERAGECOEF; i++) {
		sum += wiiNunchuk_averageAngle[i];
	}

	return (sum/WIINUNCHUK_ANGLEAVERAGECOEF);
}

#endif

void wiiNunchuk_update()
{
	uint8_t i = 0;
	uint8_t buff[WIINUNCHUK_READBYTES];

	memset(buff, 0, sizeof(buff));

	/* Request data */
	i2c_start_wait(WIINUNCHUK_ADDR | I2C_WRITE);
	i2c_write(0x00);
	i2c_stop();
	i2c_start_wait(WIINUNCHUK_ADDR | I2C_READ);

	for(i = 0; i < WIINUNCHUK_READBYTES; i++) {
		if(i == WIINUNCHUK_READBYTES - 1)
			buff[i] = i2c_readNak();
		else
			buff[i] = i2c_readAck();
	}

	i2c_stop();

	/* Decodebyte */
	#if WIINUNCHUK_DECODEBYTE == 1
	for(i = 0; i < WIINUNCHUK_READBYTES; i++) {
		buff[i] = wiiNunchuk_decode(buff[i]);
	}

	#endif

	/* Record old records */
	#if WIINUNCHUK_PULSEBUTTON == 1
	wiiNunchuk_lastButtonZ = wiiNunchuk_buttonZ;
	wiiNunchuk_lastButtonC = wiiNunchuk_buttonC;

	#endif

	/* Get joypad */
	wiiNunchuk_joyX = buff[0];
	wiiNunchuk_joyY = buff[1];

	/* Get button */
	wiiNunchuk_buttonZ = !(buff[5] & 0b00000001);
	wiiNunchuk_buttonC = !((buff[5] & 0b00000010) >> 1);

	/* Get Angle */
	wiiNunchuk_angleX = (buff[2] << 2) + ((buff[5] & (0b00000011 << (1*2)) >> (1*2)));
	wiiNunchuk_angleY = (buff[3] << 2) + ((buff[5] & (0b00000011 << (2*2)) >> (2*2)));
	wiiNunchuk_angleZ = (buff[4] << 2) + ((buff[5] & (0b00000011 << (3*2)) >> (3*2)));

	/* Filter angle */
	#if WIINUNCHUK_ANGLEFILTER == 1
	wiiNunchuk_angleX = wiiNunchuk_averageFilter(wiiNunchuk_angleX, (int *) wiiNunchuk_averageAngleX);
	wiiNunchuk_angleY = wiiNunchuk_averageFilter(wiiNunchuk_angleY, (int *) wiiNunchuk_averageAngleY);
	wiiNunchuk_angleZ = wiiNunchuk_averageFilter(wiiNunchuk_angleZ, (int *) wiiNunchuk_averageAngleZ);

	#endif
}

void wiiNunchuk_init()
{
	#if WIINUNCHUK_I2CINIT == 1
	i2c_init();
	_delay_us(100);

	#endif

	/* Standard init:    0x40 -> 0x00
	 * Alternative init: 0xF0 -> 0x55 followed by 0xFB -> 0x00, lets us use 3rd
	 * party nunchucks */
	i2c_start_wait(WIINUNCHUK_ADDR | I2C_WRITE);
	i2c_write(0xF0);
	i2c_write(0x55);
	i2c_stop();
	i2c_start_wait(WIINUNCHUK_ADDR | I2C_WRITE);
	i2c_write(0xFB);
	i2c_write(0x00);
	i2c_stop();

	wiiNunchuk_update();
}
