/*******************************************************************************
 * @file                test_Accel.c
 * @brief               Program to test accelerometer functionality
 * @date                December 4th, 2014
 * @authors             Alexander Volkov <av333@cornell.edu>,
 *                      Kevin Bonett     <kb476@cornell.edu>,
 *                      Andre Heil       <avh34@cornell.edu>
 * @version             1.0
 *
 * This program uses the TRT operating system to test the MMA8452Q
 * accelerometer. It simply reads the current accelerometer data and displays it
 * every second to serial.
 *
 ******************************************************************************/


/*** Program Macros ***********************************************************/

#define MMA8452Q_ADDR  0x3A

#define SEM_RX_ISR_SIGNAL 1
#define SEM_STRING_DONE   2

#define UART_BAUD      9600


/*** Include Files ************************************************************/

#include "../lib/TRT/trtSettings.h"
#include "../lib/TRT/trtKernel_644.c"

#include <inttypes.h>
#include <stdio.h>
#include <string.h>
#include <avr/io.h>
#include <avr/sleep.h>
#include <util/delay.h>

#include "../lib/TRT/trtUART.h"
#include "../lib/TRT/trtUART.c"

#include "../lib/I2C/I2C_lib.h"
#include "../lib/I2C/I2C_lib.c"


/*** Program Variables ********************************************************/

/* UART file descriptor */
FILE uart_str = FDEV_SETUP_STREAM(uart_putchar, uart_getchar, _FDEV_SETUP_RW);

/* Input arguments to each thread, not actually used. */
int args[1];


/*** Program Function Prototypes **********************************************/

int  main       (void);
void initialize (void);

void testTask   (void *args);


/*** Program Functions ********************************************************/

int main(void)
{
	initialize();

	/* Sets up the kernel data structures. The parameter is the desired starck
	 * size of the idle task. For a null idle task, a stack size of 80 should be
	 * sufficient. */
	trtInitKernel(80);

	/* Creates semaphores with identifier semnbr and initial values initval. */
	trtCreateSemaphore(SEM_RX_ISR_SIGNAL, 0);
	trtCreateSemaphore(SEM_STRING_DONE,   0);

	/* Identifies the three functions to the kernel as threads. The parameters
	 * specify pointers to the functions, the desired stack size, the initial
	 * release time, the initial deadline time, and an abitrary data input
	 * structure. */
	trtCreateTask(testTask, 200, SECONDS2TICKS(0.5), SECONDS2TICKS(0.5), &(args[0]));

	/* Choose our preferred sleep mode */
	set_sleep_mode(SLEEP_MODE_IDLE);

	/* Set sleep enable (SE) bit */
	sleep_enable();

	/* Sleep the CPU when a task isn't running. */
	while (1) { sleep_cpu(); }
}

void initialize(void)
{
	i2c_init();
	trt_uart_init();

	stdout = stdin = stderr = &uart_str;
	fprintf(stdout, "TinyRealTime: 2/9/09\n\r");

	/* Single byte write to set the accelerometer to active mode */
	i2c_start_wait(MMA8452Q_ADDR | I2C_WRITE);
	i2c_write(0x2A);
	i2c_write(0x01);
	i2c_stop();

}

void testTask(void *args)
{
	uint32_t rel, dead;

	int16_t outX   = 0;
	int16_t outY   = 0;
	int16_t outZ   = 0;

	while (1)
	{
		/* Start to read accelerometer data, start at 0x01, following the multi-
		 * byte protocol found on pg. 17 of the MMA8452Q data sheet */
		i2c_start_wait(MMA8452Q_ADDR | I2C_WRITE);
		i2c_write(0x01);

		i2c_rep_start(MMA8452Q_ADDR | I2C_READ);

		outX  = (i2c_readAck() << 8);
		outX |= (i2c_readAck());
		outX  = (outX >> 4);

		outY  = (i2c_readAck() << 8);
		outY |= (i2c_readAck());
		outY  = (outY >> 4);

		outZ  = (i2c_readAck() << 8);
		outZ |= (i2c_readNak());
		outZ  = (outZ >> 4);

		i2c_stop();

		/* Print results to serial */
		fprintf(stdout, "outX  : %i\n\r", outX);
		fprintf(stdout, "outY  : %i\n\r", outY);
		fprintf(stdout, "outZ  : %i\n\r", outZ);
		fprintf(stdout, "\n\r");

		rel  = trtCurrentTime() + SECONDS2TICKS(1.0);
		dead = trtCurrentTime() + SECONDS2TICKS(1.3);
		trtSleepUntil(rel, dead);
	}
}
